from typing import Optional

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.deletion import PROTECT
from edc_crf.stubs import CrfModelStub, TCrfModelStub
from edc_model.models import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import SubjectScheduleModelMixin

from edc_visit_tracking.stubs import SubjectVisitModelStub, TSubjectVisitModelStub

from ...crf_date_validator import CrfDateValidator
from ...managers import CrfModelManager
from ..visit_model_mixin import VisitModelMixin


class VisitMethodsModelMixin(models.Model):
    """A model mixin for CRFs and Requisitions"""

    @property
    def visit_code(self: CrfModelStub) -> str:
        return self.subject_visit.visit_code

    @property
    def visit(self: CrfModelStub) -> SubjectVisitModelStub:
        """Returns the model instance of the visit foreign key
        attribute.
        """
        visit = None
        fields = {field.name: field for field in self._meta.fields}
        for name, field in fields.items():
            try:
                assert field.related_model is not None
            except (AttributeError, AssertionError):
                pass
            else:
                if issubclass(field.related_model, (VisitModelMixin,)):
                    try:
                        visit = getattr(self, name)
                    except ObjectDoesNotExist:
                        pass
        return visit

    @classmethod
    def visit_model_attr(cls) -> str:
        """Returns the field name for the visit model
        foreign key.
        """
        visit_model_attr = None
        fields = {field.name: field for field in cls._meta.fields}
        for name, field in fields.items():
            try:
                assert field.related_model is not None
            except (AttributeError, AssertionError):
                pass
            else:
                if issubclass(field.related_model, (VisitModelMixin,)):
                    visit_model_attr = name
        return visit_model_attr

    @classmethod
    def visit_model(cls: TCrfModelStub) -> str:
        """Returns the visit foreign key model in
        label_lower format.
        """
        return cls.visit_model_cls()._meta.label_lower

    @classmethod
    def visit_model_cls(cls) -> Optional[TSubjectVisitModelStub]:
        """Returns the visit foreign key attribute model class."""
        fields = {field.name: field for field in cls._meta.fields}
        for name, field in fields.items():
            if name == cls.visit_model_attr():
                return field.related_model
        return None

    class Meta:
        abstract = True


class VisitTrackingCrfModelMixin(
    VisitMethodsModelMixin, SubjectScheduleModelMixin, models.Model
):
    """Base mixin for all CRF models.

    You need to define the visit model foreign_key, e.g:

        subject_visit = models.ForeignKey(SubjectVisit)

    and specify the `natural key` with its dependency of the visit model.
    """

    subject_visit = models.ForeignKey(settings.SUBJECT_VISIT_MODEL, on_delete=PROTECT)

    crf_date_validator_cls = CrfDateValidator

    report_datetime = models.DateTimeField(
        verbose_name="Report Date",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=get_utcnow,
        help_text=(
            "If reporting today, use today's date/time, otherwise use "
            "the date/time this information was reported."
        ),
    )

    objects = CrfModelManager()

    def __str__(self) -> str:
        return str(self.subject_visit)

    def natural_key(self) -> tuple:
        return tuple(
            getattr(self, self.visit_model_attr()).natural_key(),
        )

    # noinspection PyTypeHints
    natural_key.dependencies = [settings.SUBJECT_VISIT_MODEL]  # type:ignore

    @property
    def visit(self: CrfModelStub) -> SubjectVisitModelStub:
        return self.subject_visit

    @property
    def subject_identifier(self: "VisitMethodsModelMixin"):
        return self.visit.subject_identifier

    class Meta:
        abstract = True
        indexes = [models.Index(fields=["subject_visit", "report_datetime"])]



from django.contrib.sites.managers import CurrentSiteManager as BaseCurrentSiteManager
from django.db import models


class CrfModelManager(models.Manager):
    """A manager class for Crf models, models that have an FK to
    the visit model.
    """

    use_in_migrations = True

    def get_by_natural_key(
        self,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ):
        instance = self.model.visit_model_cls().objects.get_by_natural_key(
            subject_identifier,
            visit_schedule_name,
            schedule_name,
            visit_code,
            visit_code_sequence,
        )
        return self.get(**{self.model.visit_model_attr(): instance})


class VisitModelManager(models.Manager):
    """A manager class for visit models."""

    use_in_migrations = True

    def get_by_natural_key(
        self,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ):
        return self.get(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )


class CurrentSiteManager(BaseCurrentSiteManager, CrfModelManager):

    use_in_migrations = True

    pass



from typing import Union

from django.db import models
from edc_crf.stubs import CrfModelStub

from ..stubs import SubjectVisitModelStub
from ..visit_sequence import VisitSequence, VisitSequenceError


class PreviousVisitError(Exception):
    pass


class PreviousVisitModelMixin(models.Model):
    """A model mixin to force the user to complete visit model
    instances in sequence.

    * Ensures the previous visit exists before allowing save()
      by raising PreviousVisitError.
    * If the visit is the first in the sequence, save() is allowed.
    """

    visit_sequence_cls = VisitSequence

    def save(
        self: Union[CrfModelStub, SubjectVisitModelStub, "PreviousVisitModelMixin"],
        *args,
        **kwargs
    ):
        self.validate_visit_sequence()
        super().save(*args, **kwargs)  # type: ignore

    def validate_visit_sequence(self):
        try:
            appointment = self.subject_visit.appointment
        except AttributeError:
            appointment = self.appointment
        visit_sequence = self.visit_sequence_cls(appointment=appointment)
        try:
            visit_sequence.enforce_sequence()
        except VisitSequenceError as e:
            raise PreviousVisitError(e)

    @property
    def previous_visit(
        self: Union[CrfModelStub, SubjectVisitModelStub, "PreviousVisitModelMixin"]
    ):
        try:
            appointment = self.subject_visit.appointment
        except AttributeError:
            appointment = self.appointment
        visit_sequence = self.visit_sequence_cls(appointment=appointment)
        return visit_sequence.previous_visit

    class Meta:
        abstract = True



from typing import Optional

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from edc_appointment.stubs import AppointmentModelStub


class VisitSequenceError(Exception):
    pass


class VisitSequence:

    """A class that calculates the previous_visit and can enforce
    that visits are filled in sequence.
    """

    def __init__(self, appointment: AppointmentModelStub) -> None:
        self.appointment = appointment
        self.appointment_model_cls = self.appointment.__class__
        self.model_cls = getattr(
            self.appointment_model_cls,
            self.appointment_model_cls.related_visit_model_attr(),
        ).related.related_model
        self.subject_identifier = self.appointment.subject_identifier
        self.visit_schedule_name = self.appointment.visit_schedule_name
        self.visit_code = self.appointment.visit_code
        self.visit_code_sequence = self.appointment.visit_code_sequence

    def enforce_sequence(self) -> None:
        """Raises an exception if sequence is not adhered to; that is,
        the visits are not completed in order.
        """
        try:
            self.get_previous_visit()
        except ObjectDoesNotExist:
            previous_visit_code_sequence = (
                0 if not self.visit_code_sequence else self.visit_code_sequence - 1
            )
            raise VisitSequenceError(
                "Previous visit report required. Enter report for "
                f"'{self.previous_visit_code}.{previous_visit_code_sequence}' "
                f"before completing this report."
            )

    @property
    def previous_visit_code(self) -> str:
        """Return the previous visit code or the existing
        visit code if sequence is not 0.
        """
        previous_visit_code = None
        if self.visit_code_sequence != 0:
            previous_visit_code = self.visit_code
        else:
            previous = self.appointment.schedule.visits.previous(self.visit_code)
            try:
                previous_visit_code = previous.code
            except AttributeError:
                pass
        return previous_visit_code

    @property
    def previous_appointment(self) -> Optional[AppointmentModelStub]:
        """Returns the previous appointment model instance or None.

        Considers visit code sequence.

        Raises `VisitSequenceError` if the expected sequence of
        appointments is broken. Expected sequence is based on
        the visit schedule.

        """
        # TODO: consider recreating missing appointments if sequence
        #       is broken

        opts = dict(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            visit_code=self.previous_visit_code,
        )
        try:
            previous_appointment = self.appointment_model_cls.objects.get(**opts)
        except ObjectDoesNotExist as e:
            if self.previous_visit_code:
                raise VisitSequenceError(
                    f"Appointment unexpectedly does not exist. Expected "
                    f"appointment {self.previous_visit_code}. "
                    f"Got {e}"
                )
            previous_appointment = None
        except MultipleObjectsReturned:
            if self.visit_code_sequence:
                opts.update(visit_code_sequence=self.visit_code_sequence - 1)
                try:
                    previous_appointment = self.appointment_model_cls.objects.get(**opts)
                except ObjectDoesNotExist:
                    raise VisitSequenceError(
                        f"Appointment unexpectedly does not exist. Expected "
                        f"appointment for {self.previous_visit_code}."
                        f"{self.visit_code_sequence - 1}."
                    )
            else:
                opts.update(visit_code_sequence__gte=self.visit_code_sequence)
                previous_appointment = (
                    self.appointment_model_cls.objects.filter(**opts)
                    .order_by("visit_code_sequence")
                    .last()
                )
        else:
            if previous_appointment.visit_code_sequence != 0:
                raise VisitSequenceError(
                    f"Missing appointment {self.previous_visit_code}.0. "
                    f"Unexpectedly got non-zero sequence for first appointment. "
                    f"See {previous_appointment}."
                )
        return previous_appointment

    @property
    def previous_visit(self) -> Optional[AppointmentModelStub]:
        """Returns the previous visit model instance if it exists"""
        if self.previous_appointment:
            return self.model_cls.objects.get(appointment=self.previous_appointment)
        return None

    def get_previous_visit(self) -> Optional[AppointmentModelStub]:
        return self.previous_visit



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

style = color_style()

ATTR = 0
MODEL_LABEL = 1


class AppConfig(DjangoAppConfig):
    name = "edc_visit_tracking"
    verbose_name = "Edc Visit Tracking"

    # report_datetime_allowance:
    #   set to not allow CRF report_datetimes to exceed the
    #   visit report_datetime
    #   by more than X days. Set to -1 to ignore
    report_datetime_allowance: int = 30
    allow_crf_report_datetime_before_visit: bool = False
    reason_field: dict = {}

    def ready(self):
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from typing import Union

from django import forms
from django.conf import settings
from edc_crf.stubs import CrfModelFormStub

from ..crf_date_validator import (
    CrfDateValidator,
    CrfReportDateAllowanceError,
    CrfReportDateBeforeStudyStart,
    CrfReportDateIsFuture,
)
from .utils import get_subject_visit


class VisitTrackingModelFormMixin:
    """Validates subject visit and report datetime.

    Usually included in the form class declaration with
    `SubjectScheduleModelFormMixin`.
    """

    crf_date_validator_cls = CrfDateValidator
    report_datetime_allowance = getattr(settings, "DEFAULT_REPORT_DATETIME_ALLOWANCE", 0)

    def clean(self) -> dict:
        """Triggers a validation error if subject visit is None.

        If subject visit, validate report_datetime.
        """
        cleaned_data = super().clean()  # type: ignore

        self.validate_visit_tracking()

        return cleaned_data

    def validate_visit_tracking(self) -> None:
        # trigger a validation error if visit field is None
        # no comment needed since django will catch it as
        # a required field.
        if not self.subject_visit:
            if self.subject_visit_attr in self.cleaned_data:
                raise forms.ValidationError({self.subject_visit_attr: ""})
            else:
                raise forms.ValidationError(
                    f"Field `{self.subject_visit_attr}` is required (1)."
                )
        elif self.cleaned_data.get("report_datetime"):
            try:
                self.crf_date_validator_cls(
                    report_datetime_allowance=self.report_datetime_allowance,
                    report_datetime=self.cleaned_data.get("report_datetime"),
                    visit_report_datetime=self.subject_visit.report_datetime,
                )
            except (
                CrfReportDateAllowanceError,
                CrfReportDateBeforeStudyStart,
                CrfReportDateIsFuture,
            ) as e:
                raise forms.ValidationError({"report_datetime": str(e)})

    @property
    def subject_visit(self: Union["VisitTrackingModelFormMixin", CrfModelFormStub]):
        return get_subject_visit(self, subject_visit_attr=self.subject_visit_attr)

    @property
    def subject_visit_attr(self: CrfModelFormStub):
        return self._meta.model.visit_model_attr()



from django import forms
from edc_crf.stubs import CrfModelFormStub


def get_subject_visit(modelform: CrfModelFormStub, subject_visit_attr: str):
    if subject_visit_attr not in modelform.cleaned_data:
        subject_visit = getattr(modelform.instance, subject_visit_attr, None)
        if not subject_visit:
            raise forms.ValidationError(f"Field `{subject_visit_attr}` is required (2).")
    else:
        subject_visit = modelform.cleaned_data.get(subject_visit_attr)
    return subject_visit
